/*
Szymon Rusinkiewicz
Princeton University

mesh_filter.cc
Apply a variety of tranformations to a mesh
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "TriMesh.h"
#include "TriMesh_algo.h"
#include "XForm.h"

#ifndef M_PI
# define M_PI 3.14159265358979323846
#endif


// Transform the mesh by a matrix read from a file
void apply_xform(TriMesh *mesh, const char *filename)
{
	xform xf = xform::read(filename);
	int nv = mesh->vertices.size();
	for (int i = 0; i < nv; i++)
		mesh->vertices[i] = xf * mesh->vertices[i];
}


// Clip mesh to the given bounding box
bool clip(TriMesh *mesh, const char *bboxfilename)
{
	FILE *f = fopen(bboxfilename, "r");
	if (!f) {
		fprintf(stderr, "Couldn't open %s\n", bboxfilename);
		return false;
	}

	TriMesh::BBox b;
	if (fscanf(f, "%f%f%f%f%f%f",
		      &b.min[0], &b.min[1], &b.min[2],
		      &b.max[0], &b.max[1], &b.max[2]) != 6) {
		fclose(f);
		fprintf(stderr, "Couldn't read %s\n", bboxfilename);
		return false;
	}
	fclose(f);

	if (b.min[0] > b.max[0]) swap(b.min[0], b.max[0]);
	if (b.min[1] > b.max[1]) swap(b.min[1], b.max[1]);
	if (b.min[2] > b.max[2]) swap(b.min[2], b.max[2]);


	int nv = mesh->vertices.size();
	vector<bool> toremove(nv, false);
	for (int i = 0; i < nv; i++)
		if (mesh->vertices[i][0] < b.min[0] ||
		    mesh->vertices[i][0] > b.max[0] ||
		    mesh->vertices[i][1] < b.min[1] ||
		    mesh->vertices[i][1] > b.max[1] ||
		    mesh->vertices[i][2] < b.min[2] ||
		    mesh->vertices[i][2] > b.max[2])
			toremove[i] = true;

	remove_vertices(mesh, toremove);

	return true;
}


// Rotate the mesh by r around axis (ax, ay, az)
void rotate(TriMesh *mesh, float rot, const vec &axis)
{
	xform xf = xform::rot(M_PI / 180.0f * rot, axis);
	int nv = mesh->vertices.size();
	for (int i = 0; i < nv; i++)
		mesh->vertices[i] = xf * mesh->vertices[i];
}


// Translate the mesh by (tx, ty, tz)
void trans(TriMesh *mesh, const vec &transvec)
{
	int nv = mesh->vertices.size();
	for (int i = 0; i < nv; i++)
		mesh->vertices[i] += transvec;
}


// Scale the mesh by (sx, sy, sz)
void scale(TriMesh *mesh, const vec &scalevec)
{
	int nv = mesh->vertices.size();
	for (int i = 0; i < nv; i++)
		mesh->vertices[i] *= scalevec;
}


// Find center of mass
point center_of_mass(TriMesh *mesh)
{
	point com;
	float totwt = 0;

	mesh->need_faces();
	int nf = mesh->faces.size();
	for (int i = 0; i < nf; i++) {
		const point &v0 = mesh->vertices[mesh->faces[i][0]];
		const point &v1 = mesh->vertices[mesh->faces[i][1]];
		const point &v2 = mesh->vertices[mesh->faces[i][2]];

		point face_com = (v0+v1+v2) / 3.0f;
		float wt = len((v1-v0) CROSS (v2-v0));
		com += wt * face_com;
		totwt += wt;
	}
	return com / totwt;
}


// Scale the mesh so that mean squared distance from center of mass is 1
void normalize_variance(TriMesh *mesh)
{
	point com = center_of_mass(mesh);
	trans(mesh, -com);

	float var = 0;
	int nv = mesh->vertices.size();
	for (int i = 0; i < nv; i++)
		var += len2(mesh->vertices[i]);
	float s = 1.0f / sqrtf(var);
	scale(mesh, vec(s,s,s));

	trans(mesh, com);
}


void usage(const char *myname)
{
	fprintf(stderr, "Usage: %s infile [options] [outfile]\n", myname);
	fprintf(stderr, "Options: -color		Add per-vertex color\n");
	fprintf(stderr, "	 -nocolor	Remove per-vertex color\n");
	fprintf(stderr, "	 -conf		Add per-vertex confidence\n");
	fprintf(stderr, "	 -noconf	Remove per-vertex confidence\n");
	fprintf(stderr, "	 -tstrip	Convert to use triangle strips\n");
	fprintf(stderr, "	 -notstrip	Unpack triangle strips to faces\n");
	fprintf(stderr, "	 -reorder	Optimize order of vertices\n");
	fprintf(stderr, "	 -faceflip	Flip the order of vertices within each face\n");
	fprintf(stderr, "	 -edgeflip	Optimize triangle connectivity by flipping edges\n");
	fprintf(stderr, "	 -subdiv	Subdivide faces (planar)\n");
	fprintf(stderr, "	 -loop		Perform Loop subdivision\n");
	fprintf(stderr, "	 -smooth s	Smooth surface with sigma=s*edgelength\n");
	fprintf(stderr, "	 -smoothnorm s	Diffuse normals with sigma=s*edgelength\n");
	fprintf(stderr, "	 -lmsmooth n	Perform n iterations of lambda-mu smoothing\n");
	fprintf(stderr, "	 -inflate s	Create offset surface s*edgelength away\n");
	fprintf(stderr, "	 -clip bbox	Clip to the given bbox (file has 6 numbers)\n");
	fprintf(stderr, "	 -xform file.xf	Transform by the given matrix\n");
	fprintf(stderr, "	 -rot r x y z	Rotate r degrees around axis (x,y,z)\n");
	fprintf(stderr, "	 -trans x y z	Translate by (x,y,z)\n");
	fprintf(stderr, "	 -scale x y z	Scale by (x,y,z)\n");
	fprintf(stderr, "	 -bbcenter	Translate so center of bbox is at (0,0,0)\n");
	fprintf(stderr, "	 -comcenter	Translate so center of mass is at (0,0,0)\n");
	fprintf(stderr, "	 -bbnorm	Scale so bbox has maximum extent 1\n");
	fprintf(stderr, "	 -varnorm	Scale so variance (RMS distance) from center is 1\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	if (argc < 3)
		usage(argv[0]);
	const char *filename = argv[1];

	TriMesh *themesh = TriMesh::read(filename);
	if (!themesh)
		usage(argv[0]);

	bool have_tstrips = !themesh->tstrips.empty();
	for (int i = 2; i < argc; i++) {
		if (!strcmp(argv[i], "-color") ||
		    !strcmp(argv[i], "-colors")) {
			if (themesh->colors.empty()) {
				int nv = themesh->vertices.size();
				themesh->colors.resize(nv, Color::white());
			}
		} else if (!strcmp(argv[i], "-nocolor") ||
			   !strcmp(argv[i], "-nocolors")) {
			themesh->colors.clear();
		} else if (!strcmp(argv[i], "-conf")) {
			if (themesh->confidences.empty()) {
				int nv = themesh->vertices.size();
				themesh->confidences.resize(nv, 1);
			}
		} else if (!strcmp(argv[i], "-noconf")) {
			themesh->confidences.clear();
		} else if (!strcmp(argv[i], "-tstrip") ||
			   !strcmp(argv[i], "-tstrips") ||
			   !strcmp(argv[i], "-strip") ||
			   !strcmp(argv[i], "-strips")) {
			themesh->need_tstrips();
			have_tstrips = true;
		} else if (!strcmp(argv[i], "-notstrip") ||
			   !strcmp(argv[i], "-notstrips") ||
			   !strcmp(argv[i], "-nostrip") ||
			   !strcmp(argv[i], "-nostrips") ||
			   !strcmp(argv[i], "-unstrip")) {
			themesh->need_faces();
			themesh->tstrips.clear();
			have_tstrips = false;
		} else if (!strcmp(argv[i], "-reorder")) {
			reorder_verts(themesh);
		} else if (!strcmp(argv[i], "-faceflip")) {
			faceflip(themesh);
		} else if (!strcmp(argv[i], "-edgeflip")) {
			edgeflip(themesh);
		} else if (!strcmp(argv[i], "-subdiv")) {
			subdiv(themesh, false);
		} else if (!strcmp(argv[i], "-loop")) {
			subdiv(themesh);
		} else if (!strcmp(argv[i], "-smooth")) {
			i++;
			if (!(i < argc && (isdigit(argv[i][0]) || argv[i][0] == '.'))) {
				fprintf(stderr, "-smooth requires one float parameter: s\n");
				usage(argv[0]);
			}
			float amount = atof(argv[i]) * themesh->feature_size();
			smooth_mesh(themesh, amount);
			themesh->pointareas.clear();
			themesh->normals.clear();
		} else if (!strcmp(argv[i], "-smoothnorm")) {
			i++;
			if (!(i < argc && (isdigit(argv[i][0]) || argv[i][0] == '.'))) {
				fprintf(stderr, "-smoothnorm requires one float parameter: s\n");
				usage(argv[0]);
			}
			float amount = atof(argv[i]) * themesh->feature_size();
			diffuse_normals(themesh, amount);
		} else if (!strcmp(argv[i], "-lmsmooth")) {
			i++;
			if (!(i < argc && isdigit(argv[i][0]))) {
				fprintf(stderr, "-lmsmooth requires one int parameter: n\n");
				usage(argv[0]);
			}
			int niters = atoi(argv[i]);
			lmsmooth(themesh, niters);
		} else if (!strcmp(argv[i], "-inflate")) {
			i++;
			if (!(i < argc && (isdigit(argv[i][0]) || argv[i][0] == '.'))) {
				fprintf(stderr, "-inflate requires one float parameter: s\n");
				usage(argv[0]);
			}
			float amount = atof(argv[i]) * themesh->feature_size();
			inflate(themesh, amount);
		} else if (!strcmp(argv[i], "-clip")) {
			i++;
			if (!(i < argc)) {
				fprintf(stderr, "-clip requires one argument\n");
				usage(argv[0]);
			}
			if (!clip(themesh, argv[i]))
				usage(argv[0]);
		} else if (!strcmp(argv[i], "-xf") ||
			   !strcmp(argv[i], "-xform")) {
			i++;
			if (!(i < argc)) {
				fprintf(stderr, "-xform requires one argument\n");
				usage(argv[0]);
			}
			apply_xform(themesh, argv[i]);
		} else if (!strcmp(argv[i], "-rot") ||
			   !strcmp(argv[i], "-rotate")) {
			i += 4;
			if (!(i < argc)) {
				fprintf(stderr, "-rot requires four arguments\n");
				usage(argv[0]);
			}
			vec a(atof(argv[i-2]), atof(argv[i-1]), atof(argv[i]));
			rotate(themesh, atof(argv[i-3]), a);
		} else if (!strcmp(argv[i], "-trans") ||
			   !strcmp(argv[i], "-translate")) {
			i += 3;
			if (!(i < argc)) {
				fprintf(stderr, "-trans requires three arguments\n");
				usage(argv[0]);
			}
			vec t(atof(argv[i-2]), atof(argv[i-1]), atof(argv[i]));
			trans(themesh, t);
		} else if (!strcmp(argv[i], "-scale")) {
			i += 3;
			if (!(i < argc)) {
				fprintf(stderr, "-scale requires three arguments\n");
				usage(argv[0]);
			}
			vec s(atof(argv[i-2]), atof(argv[i-1]), atof(argv[i]));
			scale(themesh, s);
		} else if (!strcmp(argv[i], "-bbcenter")) {
			themesh->need_bbox();
			trans(themesh, -themesh->bbox.center());
		} else if (!strcmp(argv[i], "-avgcenter")) {
			trans(themesh, -center_of_mass(themesh));
		} else if (!strcmp(argv[i], "-bbnorm")) {
			themesh->need_bbox();
			vec l = themesh->bbox.size();
			float ll = max(max(l[0], l[1]), l[2]);
			trans(themesh, -themesh->bbox.center());
			float s = 1.0f / ll;
			scale(themesh, vec(s,s,s));
			trans(themesh, themesh->bbox.center());
		} else if (!strcmp(argv[i], "-varnorm")) {
			normalize_variance(themesh);
		} else if (i == argc - 1 &&
			   (argv[i][0] != '-' || argv[i][1] == '\0')) {
			if (have_tstrips && themesh->tstrips.empty())
				themesh->need_tstrips();
			themesh->write(argv[i]);
		} else {
			fprintf(stderr, "Unrecognized option [%s]\n", argv[i]);
			usage(argv[0]);
		}
	}
}

